#!/usr/bin/env python3
'''Main Entrypoint Module
'''
import typing
import argparse
import logging
from textwrap import dedent
import sys
from importlib.metadata import version as metadata_version


# Automatic Tab Completion
# Production Wrapper (keep the API)
class shtab:
    # Global Markers
    FILE = None
    DIRECTORY = None

    # API
    def add_argument_to(*args: typing.Any, **kwargs: typing.Any):
        pass


if __debug__:
    # Try to import the real deal
    try:
        import shtab  # type: ignore[no-redef,import-not-found] # noqa: F811
    except ImportError:
        pass


# Workaround for https://github.com/python/cpython/issues/101640
if sys.version_info >= (3, 12):
    ArgumentParser = argparse.ArgumentParser
else:
    class ArgumentParser(argparse.ArgumentParser):
        '''Wrapper for `ArgumentParser <argparse.ArgumentParser>`.

        Fixes CPython issue `#101640
        <https://github.com/python/cpython/issues/101640>`__.
        '''
        def _print_message(self, message, file=None):
            if file is None:
                pass  # Don't print if the file is unavailable
            else:
                super()._print_message(message, file=file)


# Make sure BooleanOptionalAction is available
if sys.version_info >= (3, 9):
    BooleanOptionalAction = argparse.BooleanOptionalAction
else:
    # Port from Python 3.11: https://github.com/python/cpython/blob/3.11/Lib/argparse.py#L885-L921
    class BooleanOptionalAction(argparse.Action):
        '''Port `BooleanOptionalAction
        <https://docs.python.org/3.11/library/argparse.html#action>`__ from
        Python 3.11.
        '''
        def __init__(self,
                     option_strings,
                     dest,
                     default=None,
                     required=False,
                     help=None,
                     ):

            _option_strings = []
            for option_string in option_strings:
                _option_strings.append(option_string)
                if option_string.startswith('--'):
                    option_string = '--no-' + option_string[2:]
                    _option_strings.append(option_string)

            super().__init__(
                option_strings=_option_strings,
                dest=dest,
                nargs=0,
                default=default,
                type=type,
                required=required,
                help=help,
            )

        def __call__(self, parser, namespace, values, option_string=None):
            if option_string in self.option_strings:
                setattr(namespace, self.dest, not option_string.startswith('--no-'))

        def format_usage(self):
            return ' | '.join(self.option_strings)


logger = logging.getLogger(__name__)


def main(PROJECT_NAME, PROJECT_VERSION,
         ):
    '''
    Main entrypoint to be configured
    '''
    parser = ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=dedent('''
        Prints Hello World
        '''),
        epilog=dedent(f'''
        Version {PROJECT_VERSION}
        '''),
    )
    # Automatic Tab Completion
    # - Mark certain arguments with:
    #   - `parser.add_argument(...).complete = shtab.FILE`: Complete file names
    #   - `parser.add_argument(...).complete = shtab.DIRECTORY`: Complete directory names
    shtab.add_argument_to(parser, '--generate-shtab-completion', help=argparse.SUPPRESS)

    parser.add_argument('--version', action='version', version=PROJECT_VERSION)
    parser.add_argument('-v', '--verbose', dest='loglevel',
                        action='store_const', const=logging.DEBUG, default=logging.INFO,
                        help='Add more details to the standard error log')

    args = parser.parse_args()

    # Logs
    logs_fmt = '%(levelname)-5.5s %(name)s@%(funcName)s %(message)s'
    try:
        import coloredlogs  # type: ignore
        coloredlogs.install(level=args.loglevel, fmt=logs_fmt)
    except ImportError:
        logging.basicConfig(level=args.loglevel, format=logs_fmt)
    logging.captureWarnings(True)

    print('Implement a CLI interface')

    return 0


# Release Process
def entrypoint():
    '''
    Entrypoint for executable
    '''
    __project_name__ = __package__
    __version__ = metadata_version(__project_name__)
    return main(
        PROJECT_NAME=__project_name__,
        PROJECT_VERSION=__version__,
    )


if __name__ == "__main__":
    sys.exit(entrypoint())



#!/usr/bin/env python3
import logging
from dataclasses import dataclass
import typing

import hashlib
import pickle
import hmac
import lzma


logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class Digest:
    name: str
    fn: typing.Callable[[], typing.Any]
    size: int


@dataclass(frozen=True)
class CoDec:
    name: str
    compress: typing.Callable
    decompress: typing.Callable


DEFAULT_DIGEST = Digest(
    'BLAKE2b, with 64 bytes',
    lambda: hashlib.blake2b(digest_size=64),
    size=64,
)
# TODO: Abandon LZMA, use Zstandard (pyzstd)
CODEC_LZMA = CoDec(
    'LZMA in XZ, with SHA256 checksums',
    compress=lambda fobj: lzma.open(fobj, 'wb', format=lzma.FORMAT_XZ, check=lzma.CHECK_SHA256),
    decompress=lambda fobj: lzma.open(fobj, 'rb', format=lzma.FORMAT_XZ),
)


def dump(obj, file, *, hmac_key: bytes,
         pickle_protocol: int = pickle.HIGHEST_PROTOCOL,
         digest: Digest = DEFAULT_DIGEST,
         codec: CoDec = CODEC_LZMA) -> None:
    logger.debug('# Saving Object')
    logger.debug('- HMAC Digest: %s', digest.name)
    logger.debug('-       CoDec: %s', codec.name)
    # # Pickle
    obj_pickled = pickle.dumps(obj, protocol=pickle_protocol)
    logger.debug('Pickle: %d bytes', len(obj_pickled))
    # # HMAC
    hmac_obj = hmac.new(hmac_key, digestmod=digest.fn)
    hmac_obj.update(obj_pickled)
    obj_hmac = hmac_obj.digest()
    logger.debug('HMAC [%s]: "0x%s"', hmac_obj.name, hmac_obj.hexdigest())
    # # Compress
    size_uncompressed = 0
    if file.seekable():
        tell_before = file.tell()
    else:
        tell_before = None
    # Write HMAC (uncompressed)
    size_uncompressed += file.write(obj_hmac)
    with codec.compress(file) as compress_obj:
        # Write Pickled Object (compressed)
        size_uncompressed += compress_obj.write(obj_pickled)
    # TODO: Use Zstandard
    # | See https://github.com/facebook/zstd/blob/release/doc/zstd_compression_format.md#skippable-frames
    # |     to include the HMAC as a skippable frame (not as uncompressed data)
    logger.debug('| Size HMAC: %d bytes', digest.size)
    logger.debug('| Size Uncompressed: %d bytes', size_uncompressed)
    if tell_before is not None:
        size_compressed = file.tell() - tell_before
        logger.debug('| Size Compressed: %d bytes', size_compressed)
        compress_ratio = size_compressed / size_uncompressed
        logger.debug('| Compression Ratio: %.2f%%', compress_ratio * 100)


def dumps(obj, *, hmac_key: bytes,
          pickle_protocol: int = pickle.HIGHEST_PROTOCOL,
          digest: Digest = DEFAULT_DIGEST,
          codec: CoDec = CODEC_LZMA) -> None:
    # TODO: Move `dump` in here, switch the implementation around
    raise NotImplementedError


def load(file, *, hmac_key: bytes,
         digest: Digest = DEFAULT_DIGEST,
         codec: CoDec = CODEC_LZMA) -> typing.Any:
    logger.debug('# Loading Object')
    logger.debug('- HMAC Digest: %s', digest.name)
    logger.debug('-       CoDec: %s', codec.name)
    if file.seekable():
        tell_before = file.tell()
    else:
        tell_before = None
    # Read HMAC
    logger.debug('| Size HMAC: %d bytes', digest.size)
    hmac_at_header = file.read(digest.size)
    assert len(hmac_at_header) == digest.size
    # Read Compressed Pickle
    pickle_bytes = None
    with codec.decompress(file) as uncompress_obj:
        pickle_bytes = uncompress_obj.read()
    assert pickle_bytes is not None
    size_uncompressed = len(pickle_bytes) + digest.size
    logger.debug('| Size Uncompressed: %d bytes', size_uncompressed)
    if tell_before is not None:
        size_compressed = file.tell() - tell_before
        logger.debug('| Size Compressed: %d bytes', size_compressed)
        compress_ratio = size_compressed / size_uncompressed
        logger.debug('| Compression Ratio: %.2f%%', compress_ratio * 100)

    # # HMAC
    hmac_obj = hmac.new(hmac_key, digestmod=digest.fn)
    hmac_obj.update(pickle_bytes)
    hmac_at_pickle = hmac_obj.digest()
    if not hmac.compare_digest(hmac_at_pickle, hmac_at_header):
        raise ValueError('Invalid HMAC')  # TODO: raise a custom exception
    logger.debug('HMAC [%s]: "0x%s"', hmac_obj.name, hmac_obj.hexdigest())
    logger.debug('Pickle: %d bytes', len(pickle_bytes))
    # Valid, read the pickle
    return pickle.loads(pickle_bytes)


def loads(data, *, hmac_key: bytes,
          digest: Digest = DEFAULT_DIGEST,
          codec: CoDec = CODEC_LZMA) -> typing.Any:
    # TODO: Move `load` in here, switch the implementation around
    raise NotImplementedError


__all__ = [
    # Classes
    'Digest', 'CoDec',
    # Default Values
    'DEFAULT_DIGEST', 'CODEC_LZMA',
    # Functions
    'dump', 'dumps',
    'load', 'loads',
]
